import pathlib
import tempfile
import zipfile


def retrieve_data(zip_file):
    """Extract contents of data zip file and return data files
    """
    zpath = pathlib.Path(__file__).resolve().parent / "data" / zip_file
    # unpack
    arc = zipfile.ZipFile(str(zpath))

    # extract all files to a temporary directory
    edest = tempfile.mkdtemp(prefix=zpath.name)
    arc.extractall(edest)
    return pathlib.Path(edest)



import pathlib
from PyQt6 import uic, QtCore, QtWidgets
import pkg_resources


from ..util import is_dir_writable


class Preferences(QtWidgets.QDialog):
    """Preferences dialog"""
    feature_changed = QtCore.pyqtSignal()

    def __init__(self, parent, *args, **kwargs):
        QtWidgets.QWidget.__init__(self, parent=parent, *args, **kwargs)
        path_ui = pkg_resources.resource_filename(
            "mpl_data_cast.gui", "preferences.ui")
        uic.loadUi(path_ui, self)
        self.settings = QtCore.QSettings()
        self.parent = parent

        #: configuration keys, corresponding widgets, and defaults
        self.config_pairs = [
            ["rtdc/output_path", self.rtdc_output_path,
             pathlib.Path.cwd()],
            ["rtdc/tree_depth_limit", self.tree_depth_limit, 3],
        ]
        self.reload()

        # signals
        self.btn_apply = self.buttonBox.button(
            QtWidgets.QDialogButtonBox.StandardButton.Apply)
        self.btn_apply.clicked.connect(self.on_settings_apply)
        self.btn_cancel = self.buttonBox.button(
            QtWidgets.QDialogButtonBox.StandardButton.Cancel)
        self.btn_restore = self.buttonBox.button(
            QtWidgets.QDialogButtonBox.StandardButton.RestoreDefaults)
        self.btn_restore.clicked.connect(self.on_settings_restore)
        self.select_rtdc_output_path.clicked.connect(
            self.on_task_select_output_dir)
        # tab changed
        self.tabWidget.currentChanged.connect(self.on_tab_changed)

    def reload(self):
        """Read configuration or set default parameters"""
        for key, widget, default in self.config_pairs:
            value = self.settings.value(key, default)
            if isinstance(widget, QtWidgets.QCheckBox):
                widget.setChecked(bool(int(value)))
            elif isinstance(widget, QtWidgets.QLineEdit):
                widget.setText(str(value))
            elif isinstance(widget, QtWidgets.QSpinBox):
                widget.setValue(int(value))
            else:
                raise NotImplementedError("No rule for '{}'".format(key))

    @QtCore.pyqtSlot()
    def on_settings_apply(self):
        """Save current changes made in UI to settings and reload UI"""
        for key, widget, default in self.config_pairs:
            if isinstance(widget, QtWidgets.QCheckBox):
                value = int(widget.isChecked())
            elif isinstance(widget, QtWidgets.QLineEdit):
                value = widget.text().strip()
                if widget is self.rtdc_output_path:
                    prev_value = self.settings.value(key)
                    path = pathlib.Path(widget.text().strip())
                    if path.exists():
                        value = str(path)
                    else:
                        msg = QtWidgets.QMessageBox()
                        msg.setIcon(QtWidgets.QMessageBox.Icon.Warning)
                        msg.setText("Given path does not exist! Please "
                                    "check!\nPrevious setting will be"
                                    " restored.")
                        msg.setWindowTitle("Warning")
                        msg.exec()
                        if pathlib.Path(prev_value).exists():
                            value = prev_value
                        else:
                            msg = QtWidgets.QMessageBox()
                            msg.setIcon(QtWidgets.QMessageBox.Icon.Warning)
                            msg.setText("Path from previous settings does not "
                                        "exist either, original default output"
                                        " path is restored.")
                            msg.setWindowTitle("Warning")
                            msg.exec()
                            value = pathlib.Path.cwd()
            elif isinstance(widget, QtWidgets.QSpinBox):
                value = int(widget.value())
            else:
                raise NotImplementedError("No rule for '{}'".format(key))
            self.settings.setValue(key, value)

        # reload UI to give visual feedback
        self.reload()

    @QtCore.pyqtSlot()
    def on_settings_restore(self):
        self.settings.clear()
        self.reload()

    @QtCore.pyqtSlot()
    def on_tab_changed(self):
        self.btn_apply.setEnabled(True)
        self.btn_cancel.setEnabled(True)
        self.btn_restore.setEnabled(True)

    @QtCore.pyqtSlot()
    def on_task_select_output_dir(self) -> None:
        p = QtWidgets.QFileDialog.getExistingDirectory(
            parent=self,
            caption="Select output directory:",
            directory=self.rtdc_output_path.text())
        if p:
            if is_dir_writable(p):
                self.rtdc_output_path.setText(str(p))
            else:
                raise ValueError(f"Directory {p} is not writable!")



from PyQt6 import QtWidgets, QtCore, uic
import pkg_resources
import pathlib

from ..path_tree import PathTree, list_items_in_tree


class InputWidget(QtWidgets.QWidget):
    """Widget in the RTDC tab view dealing with the input directory.
    Contains a lineEdit, a button, and a treeview widget."""
    def __init__(self, *args, **kwargs):
        super(InputWidget, self).__init__(*args, **kwargs)

        path_ui = pkg_resources.resource_filename("mpl_data_cast.gui",
                                                  "widget_input.ui")
        uic.loadUi(path_ui, self)

        self.path = None
        self.p_tree = None
        self.tree_depth_limit = 3

        self.pushButton_input_dir.clicked.connect(
            self.on_task_select_input_dir)
        self.lineEdit_input.editingFinished.connect(
            self.update_input_dir_from_lineedit)

    @QtCore.pyqtSlot()
    def on_task_select_input_dir(self) -> None:
        p = QtWidgets.QFileDialog.getExistingDirectory(
            parent=self,
            caption="Select input directory:",
            directory=str(self.path) if self.path else None)
        if p:
            self.update_input_dir(p)

    def dragEnterEvent(self, e) -> None:
        """Whether files are accepted"""
        if e.mimeData().hasUrls():
            e.accept()
        else:
            e.ignore()

    def dropEvent(self, e) -> None:
        """Add dropped directory to treeview and lineedit."""
        urls = e.mimeData().urls()
        file_1 = urls[0]
        pp = pathlib.Path(file_1.toLocalFile())
        if pp.is_dir():
            path_input = pp
        else:
            path_input = pp.parent

        self.update_input_dir(path_input)

    @QtCore.pyqtSlot()
    def update_input_dir_from_lineedit(self) -> None:
        """Executed when the input path was manually edited by the user."""
        output_dir = self.lineEdit_input.text()
        if output_dir:
            self.update_input_dir(output_dir)

    def update_input_dir(self, input_dir: str | pathlib.Path) -> None:
        """Checks if the input directory as given by the user exists and
        updates the lineEdit widget accordingly.

        Parameter
        ---------
        input_dir: str or pathlib.Path
            The directory for the input.
        """
        input_dir = pathlib.Path(input_dir)
        if input_dir.exists():
            self.path = input_dir
            self.lineEdit_input.setText(str(input_dir))
            self.update_tree()
        else:
            msg_txt = "The input directory is not valid, it does not seem" \
                      " to exist."
            msg = QtWidgets.QMessageBox(self)
            msg.setIcon(QtWidgets.QMessageBox.Icon.Warning)
            msg.setText(msg_txt)
            msg.setWindowTitle("Warning")
            msg.exec()

    def update_tree(self) -> None:
        """Update the `PathTree` object based on the current root path in
        `self.path` and update the GUI to show the new tree."""
        self.p_tree = PathTree(self.path, self.tree_depth_limit)
        self.treeWidget_input.clear()
        self.treeWidget_input.setColumnCount(self.p_tree.tree_depth + 1)

        list_items_in_tree(self.p_tree,
                           self.treeWidget_input,
                           h_level=0,
                           depth_limit=self.tree_depth_limit)

        QtWidgets.QApplication.processEvents(
            QtCore.QEventLoop.ProcessEventsFlag.AllEvents, 300)



import pathlib
import tempfile

from mpl_data_cast.mod_recipes import CatchAllRecipe


def test_basic():
    tmpdir = pathlib.Path(tempfile.mkdtemp(prefix="rcp_catchall"))
    one = tmpdir / "input" / "folder" / "a.txt"
    one.parent.mkdir(parents=True)
    one.write_text("hello testing world")
    two = tmpdir / "input" / "1.txt"
    two.write_text("Another file!")

    target = tmpdir / "output"
    target.mkdir(parents=True)

    rcp = CatchAllRecipe(path_raw=tmpdir/"input", path_tar=tmpdir/"output")
    rcp.cast()

    assert (target / "folder" / "a.txt").read_text() == "hello testing world"
    assert (target / "1.txt").read_text() == "Another file!"



def main():
    import os
    import pkg_resources
    import sys

    from PyQt6.QtWidgets import QApplication

    app = QApplication(sys.argv)
    imdir = pkg_resources.resource_filename("mpl_data_cast", "img")

    from PyQt6 import QtGui
    from mpl_data_cast.gui import MPLDataCast

    # Set Application Icon
    icon_path = os.path.join(imdir, "mpldc_icon.png")
    app.setWindowIcon(QtGui.QIcon(icon_path))

    window = MPLDataCast()

    sys.exit(app.exec())


if __name__ == "__main__":
    main()



import dclab
import dclab.cli
import hdf5plugin

from ..recipe import Recipe


class RTDCRecipe(Recipe):
    """Compress RT-DC data and append SoftwareSettings.ini to logs"""
    def convert_dataset(self, path_list, temp_path, **kwargs):
        """Compress the dataset using dclab and append SoftwareSettings.ini"""
        # first compress the .rtdc file
        dclab.cli.compress(path_out=temp_path, path_in=path_list[0])
        # the rest of the files should be log files
        cmp_kw = hdf5plugin.Zstd(clevel=5)
        if len(path_list) > 1:
            with dclab.RTDCWriter(temp_path, compression_kwargs=cmp_kw) as hw:
                for pp in path_list[1:]:
                    lines = pp.read_text().split("\n")
                    lines = [ll.rstrip() for ll in lines]
                    hw.store_log(pp.name, lines)

    def get_raw_data_iterator(self):
        """Get list of .rtdc files including associated files"""
        for pp in self.path_raw.rglob("*.rtdc"):
            path_list = [pp]
            # search for matching SoftwareSettings.ini files
            if pp.name.startswith("M"):
                pini = pp.with_name(
                    pp.name.split("_")[0]+"_SoftwareSettings.ini")
                if pini.exists():
                    path_list.append(pini)
            yield path_list



from PyQt6 import QtCore, QtTest, QtWidgets
import pathlib
import tempfile as tf
import shutil

from mpl_data_cast.gui.main import MPLDataCast
from helper import retrieve_data

data_path = pathlib.Path(__file__).resolve().parent / "data"


def test_setting_paths(qtbot, tmp_path):
    """Check that setting the input and output paths works."""
    mw = MPLDataCast()
    qtbot.addWidget(mw)
    QtWidgets.QApplication.setActiveWindow(mw)
    QtTest.QTest.qWait(100)
    path_in = retrieve_data("rcp_rtdc_mask-contour_2018.zip")
    mw.widget_input.lineEdit_input.setText(str(path_in))
    mw.widget_output.lineEdit_output.setText(str(tmp_path))
    mw.widget_input.update_input_dir_from_lineedit()
    mw.widget_output.update_output_dir_from_lineedit()
    assert str(mw.widget_output.path) == str(tmp_path)
    assert mw.widget_input.p_tree.tree_depth == 1
    assert mw.widget_output.p_tree.tree_depth == 1
    mw.close()
    QtTest.QTest.qWait(100)
    QtWidgets.QApplication.processEvents(
        QtCore.QEventLoop.ProcessEventsFlag.AllEvents, 5000)


def test_transfer_data_simple(qtbot, tmp_path, monkeypatch):
    """Check that data is transfered when clicking on the button 'Transfer'."""
    mw = MPLDataCast()
    qtbot.addWidget(mw)
    QtWidgets.QApplication.setActiveWindow(mw)
    QtTest.QTest.qWait(100)

    data = retrieve_data("rcp_rtdc_mask-contour_2018.zip")
    test_file = data / "M001_data.rtdc"
    test_file_dst = tmp_path / "M001_data.rtdc"
    shutil.copy(test_file, test_file_dst)
    tmp_dir = tf.TemporaryDirectory()
    path_out = pathlib.Path(tmp_dir.name)
    assert tmp_path.exists()
    assert path_out.exists()

    mw.widget_input.update_input_dir(tmp_path)
    mw.widget_output.update_output_dir(path_out)

    bt = mw.pushButton_transfer

    # Monkeypatch message box to always return OK
    monkeypatch.setattr(QtWidgets.QMessageBox,
                        "information",
                        lambda *args: QtWidgets.QMessageBox.StandardButton.Ok)
    qtbot.mouseClick(bt, QtCore.Qt.MouseButton.LeftButton)

    assert (path_out / "M001_data.rtdc").exists()
    tmp_dir.cleanup()
    mw.close()
    QtTest.QTest.qWait(100)
    QtWidgets.QApplication.processEvents(
        QtCore.QEventLoop.ProcessEventsFlag.AllEvents, 5000)


def test_transfer_data_advanced(qtbot, tmp_path, monkeypatch):
    """Check that data is transfered when clicking on the button 'Transfer'.
    Now includes subdirectories"""
    mw = MPLDataCast()
    qtbot.addWidget(mw)
    QtWidgets.QApplication.setActiveWindow(mw)
    QtTest.QTest.qWait(100)

    mw.settings.clear()

    data = retrieve_data("rcp_rtdc_mask-contour_2018.zip")
    test_file = data / "M001_data.rtdc"
    path_in = tf.mkdtemp(prefix="input", dir=tmp_path)
    path_in = pathlib.Path(path_in)
    path_out = tf.mkdtemp(prefix="output", dir=tmp_path)
    path_out = pathlib.Path(path_out)

    tmp_dir1 = pathlib.Path(tf.mkdtemp(dir=str(path_in)))
    tmp_dir2 = pathlib.Path(tf.mkdtemp(dir=str(tmp_dir1)))

    tmp_file1 = tmp_dir1 / "m1.rtdc"
    tmp_file2 = tmp_dir2 / "m2.rtdc"
    shutil.copy(test_file, tmp_file1)
    shutil.copy(test_file, tmp_file2)

    assert path_in.exists()
    assert path_out.exists()
    assert tmp_file1.exists()
    assert tmp_file2.exists()
    assert len(list(path_out.rglob("*.rtdc"))) == 0

    mw.widget_input.update_input_dir(path_in)
    mw.widget_output.update_output_dir(path_out)
    assert mw.widget_output.p_tree.tree_depth == 1

    bt = mw.pushButton_transfer

    # Monkeypatch message box to always return OK
    monkeypatch.setattr(QtWidgets.QMessageBox,
                        "information",
                        lambda *args: QtWidgets.QMessageBox.StandardButton.Ok)
    qtbot.mouseClick(bt, QtCore.Qt.MouseButton.LeftButton)

    assert tmp_file1.exists()
    assert tmp_file2.exists()
    assert mw.widget_output.p_tree.tree_depth == 3
    assert mw.widget_input.treeWidget_input.columnCount() == 4
    assert mw.widget_output.treeWidget_output.columnCount() == 4
    mw.close()
    QtTest.QTest.qWait(100)
    QtWidgets.QApplication.processEvents(
        QtCore.QEventLoop.ProcessEventsFlag.AllEvents, 5000)



import platform

pyi_splash = None
if platform.system() in ["Linux", "Windows"]:
    # Don't even try this on macOS.
    try:
        import pyi_splash
    except (ImportError, KeyError):
        pass


def splash_close():
    """Close the splash screen"""
    if pyi_splash is not None:
        if pyi_splash.is_alive():
            pyi_splash.close()



"""Utility methods"""
import functools
import hashlib
import pathlib
from typing import Callable


def hashfile(fname: str | pathlib.Path,
             blocksize: int = 65536,
             count: int = 0,
             constructor: Callable = hashlib.md5) -> str:
    """Compute md5 hex-hash of a file

    Parameters
    ----------
    fname: str or pathlib.Path
        path to the file
    blocksize: int
        block size in bytes read from the file
        (set to `0` to hash the entire file)
    count: int
        number of blocks read from the file
    constructor: callable
        hash algorithm constructor
    """
    path = pathlib.Path(fname).resolve()
    path_stat = path.stat()
    return _hashfile_cached(
        path=path,
        path_stats=(path_stat.st_mtime_ns, path_stat.st_size),
        blocksize=blocksize,
        count=count,
        constructor=constructor
    )


@functools.lru_cache(maxsize=100)
def _hashfile_cached(path: pathlib.Path,
                     path_stats: tuple,
                     blocksize: int = 65536,
                     count: int = 0,
                     constructor: Callable = hashlib.md5) -> str:
    """Cached hashfile using stat tuple as cache

    This is a privat function. Please use `hashfile` instead!

    Parameters
    ----------
    path: pathlib.Path
        path to the file to be hashed
    path_stats: tuple
        tuple that contains information about the size and the
        modification time of `path`. This must be specified,
        so that caching of the result is done properly in case the user
        modified `path` (this function is wrapped with
        functools.lru_cache)
    blocksize: int
        block size in bytes read from the file
        (set to `0` to hash the entire file)
    count: int
        number of blocks read from the file
    constructor: callable
        hash algorithm constructor
    """
    assert path_stats, "We need stat for validating the cache"
    hasher = constructor()
    with path.open('rb') as fd:
        buf = fd.read(blocksize)
        ii = 0
        while len(buf) > 0:
            hasher.update(buf)
            buf = fd.read(blocksize)
            ii += 1
            if count and ii == count:
                break
    return hasher.hexdigest()


def is_dir_writable(path):
    """Check whether a directory is writable

    We could play around with os.access, but ultimately there will
    always be some use case where that does not work. So we just try
    to create a file in that directory and be done with it.
    """
    path = pathlib.Path(path)
    if not path.exists():
        writable = False
    elif path.is_file():
        writable = False
    else:
        test_file = path / ".write_test~"
        try:
            test_file.touch()
        except BaseException:
            writable = False
        else:
            writable = True
        finally:
            test_file.unlink(missing_ok=True)
    return writable



import hashlib
from abc import ABC, abstractmethod
import atexit
import pathlib
import shutil
import tempfile
import traceback
import uuid
from typing import Type, Callable, List

from .util import hashfile


#: Files that are not copied (unless specified explicitly by a recipe)
IGNORED_FILE_NAMES = [
    ".DS_Store",
    "._.DS_Store",
    "Thumbs.db",
]


class Recipe(ABC):
    #: Ignored files as specified by the recipe (an addition
    #: to `IGNORED_FILE_NAMES`)
    ignored_file_names: List[str] = []

    def __init__(self,
                 path_raw: str | pathlib.Path,
                 path_tar: str | pathlib.Path):
        """Base class recipe for data conversion

        Parameters
        ----------
        path_raw: str or pathlib.Path
            Directory tree containing raw experimental data
        path_tar: str or pathlib.Path
            Target directory for converted data
        """
        #: The dataset format (defined by class name)
        self.format = self.__class__.__name__
        #: Path to raw data tree
        self.path_raw = pathlib.Path(path_raw)
        #: path to target data tree
        self.path_tar = pathlib.Path(path_tar)
        if not self.path_raw.exists():
            raise ValueError(f"Raw data path '{self.path_raw}' doesn't exist!")
        #: Temporary directory (will be deleted upon application exit)
        self.tempdir = pathlib.Path(tempfile.mkdtemp(prefix="MPL-Data-Cast_"))
        atexit.register(shutil.rmtree, self.tempdir, ignore_errors=True)

    def cast(self, path_callback: Callable = None, **kwargs) -> dict:
        """Cast the entire data tree to the target directory

        Parameters
        ----------
        path_callback: Callable
            Callable function accepting a path; used for tracking
            the progress (e.g. via the CLI)

        Returns
        -------
        result: dict
            Results dictionary with keys "success" (bool) and "errors"
            (list of tuples (path, formatted traceback))
        """
        errors = []
        # TODO: use more efficient tree structure to keep track of known files
        known_files = []
        # Copy the raw data specified by the recipe
        ds_iterator = self.get_raw_data_iterator()
        for path_list in ds_iterator:
            known_files += path_list
            if path_callback is not None:
                path_callback(path_list[0])
            targ_path = self.get_target_path(path_list)
            temp_path = self.get_temp_path(path_list)
            try:
                self.convert_dataset(path_list=path_list, temp_path=temp_path,
                                     **kwargs)
            except BaseException:
                errors.append((path_list[0], traceback.format_exc()))
                continue
            ok = self.transfer_to_target_path(temp_path=temp_path,
                                              target_path=targ_path,
                                              delete_after=True,  # [sic!]
                                              )
            if not ok:
                raise ValueError(f"Transfer to {targ_path} failed!")
        # Walk the directory tree and copy any other files
        ignored = IGNORED_FILE_NAMES + self.ignored_file_names
        for pp in self.path_raw.rglob("*"):
            if pp.is_dir() or pp.name in ignored:
                continue
            elif pp in known_files:  # this might be slow
                continue
            else:
                if path_callback is not None:
                    path_callback(pp)
                prel = pp.relative_to(self.path_raw)
                target_path = self.path_tar / prel
                target_path.parent.mkdir(parents=True, exist_ok=True)
                ok = self.transfer_to_target_path(temp_path=pp,
                                                  target_path=target_path,
                                                  delete_after=False,  # [sic!]
                                                  )
                if not ok:
                    raise ValueError(f"Transfer to {target_path} failed!")

        return {
            "success": not bool(errors),
            "errors": errors,
        }

    @abstractmethod
    def convert_dataset(self, path_list, temp_path, **kwargs):
        """Implement in subclass to do conversion"""

    @abstractmethod
    def get_raw_data_iterator(self):
        """Return list of lists of raw data paths

        Returns
        -------
        raw_data_iter: iterable of lists
            iterator (yielding lists of pathlib.Path) of which
            each item contains all files that belong to one dataset.
        """

    def get_target_path(self, path_list: list) -> pathlib.Path:
        """Get the target path for a path_list

        The target path is computed such that these relative paths
        are the same:

        - self.path_raw - path_list[0]
        - self.path_tar - target_path

        Parameters
        ----------
        path_list: list of pathlib.Path
            the input paths corresponding to a dataset

        Returns
        -------
        target_path: pathlib.Path
            the output path
        """
        prel = path_list[0].relative_to(self.path_raw)
        target_path = self.path_tar / prel
        return target_path

    def get_temp_path(self, path_list: list) -> pathlib.Path:
        """Return a unique temporary file name"""
        self.tempdir.mkdir(parents=True, exist_ok=True)
        # Create a hash of the file path
        hash1 = hashlib.md5(str(path_list[0]).encode("utf-8")).hexdigest()
        return self.tempdir / f"{hash1}_{uuid.uuid4()}_{path_list[0].name}"

    @staticmethod
    def transfer_to_target_path(temp_path: pathlib.Path,
                                target_path: pathlib.Path,
                                check_existing: bool = True,
                                delete_after: bool = False,
                                ) -> bool:
        """Transfer a file to another location

        Parameters
        ----------
        temp_path: pathlib.Path
            input file to be transferred
        target_path: pathlib.Path
            target location of the output file (including file name)
        check_existing: bool
            if `target_path` already exists, perform an MD5sum check
            and re-copy the file if the check fails
        delete_after: bool
            whether to delete `temp_path` after transfer

        Returns
        -------
        success: bool
            whether everything went as planned
        """
        target_path.parent.mkdir(parents=True, exist_ok=True)
        # compute md5hash of temp_path
        hash_ok = hashfile(temp_path)
        if target_path.exists():
            if check_existing:
                # first check the size, then the hash
                if (target_path.stat().st_size != temp_path.stat().st_size
                        or hashfile(target_path) != hash_ok):
                    # The file is not the same, delete it and try again.
                    target_path.unlink()
                    success = Recipe.transfer_to_target_path(
                        temp_path=temp_path,
                        target_path=target_path,
                        check_existing=False,
                        delete_after=False,  # [sic!]
                    )
                else:
                    # The file is the same, everything is good.
                    success = True
            else:
                # We don't know whether the file is the same, but
                # we don't care.
                success = True
        else:
            # transfer to target_path
            shutil.copy2(temp_path, target_path)
            # compute md5hash of target path
            hash_cp = hashfile(target_path)
            # compare md5hashes (verification)
            success = hash_ok == hash_cp
        if success and delete_after:
            temp_path.unlink(missing_ok=True)
        return success


def get_available_recipe_names() -> list[str]:
    names = []
    for cls in Recipe.__subclasses__():
        names.append(map_class_to_recipe_name(cls))
    return sorted(names)


def map_class_to_recipe_name(cls: Type[Recipe]) -> str:
    cls_name = cls.__name__
    assert cls_name.endswith("Recipe")
    return cls_name[:-6]


def map_recipe_name_to_class(recipe_name: str) -> Type[Recipe]:
    for cls in Recipe.__subclasses__():
        if cls.__name__.lower() == recipe_name.lower() + "recipe":
            return cls
    else:
        raise KeyError(f"Could not find class recipe for '{recipe_name}'!")



import pathlib
import h5py
import numpy as np

from ..helper import is_valid_h5_file, h5_file_contains
from ..recipe import Recipe


class OAHRecipe(Recipe):
    """Matlab file format (TopogMap.mat) for DHM data"""

    def convert_dataset(self, path_list: list, temp_path: pathlib.Path,
                        wavelength: float = None,
                        pixel_size: float = None,
                        medium_index: float = None,
                        ):
        """Convert DHM .mat data to qpformat HDF5 format"""
        with h5py.File(path_list[0]) as mat, h5py.File(temp_path, "w") as h5:
            topog = mat["topogMap"][:]
            # make sure we have a 3D stack
            if len(topog.shape) == 2:
                topog.reshape(1, topog.shape[0], -1)
            for ii in range(topog.shape[0]):
                ds = h5.create_dataset(
                    name=str(ii),
                    chunks=topog[ii].shape,
                    data=topog[ii],
                    fletcher32=True,
                )
                # QPI metadata
                ds.attrs["numerical aperture"] = mat["NA"][:].item()

                ds.attrs["wavelength"] = \
                    wavelength or mat["lambda"][:].item() * 1e-6

                ds.attrs["pos x"] = mat["positionVal"][0].item() * 1e-6
                ds.attrs["pos y"] = mat["positionVal"][1].item() * 1e-6
                ds.attrs["focus"] = mat["positionVal"][2].item() * 1e-6

                ds.attrs["pixel size"] = \
                    pixel_size or mat["res"][:].item() * 1e-6

                if "mediumIndex" in mat:  # TODO: get correct key
                    ds.attrs["medium index"] = mat["mediumIndex"].item()
                elif medium_index:
                    ds.attrs["medium index"] = medium_index

                if "frameRate" in mat:
                    dt = 1 / mat["frameRate"][:].item()
                    ds.attrs["time"] = ii * dt
                # Create and Set image attributes:
                # HDFView recognizes this as a series of images.
                # Use np.string_ as per
                # http://docs.h5py.org/en/stable/strings.html#compatibility
                ds.attrs.create('CLASS', np.string_('IMAGE'))
                ds.attrs.create('IMAGE_VERSION', np.string_('1.2'))
                ds.attrs.create('IMAGE_SUBCLASS',
                                np.string_('IMAGE_GRAYSCALE'))

            # write qpformat metadata identifier
            h5.attrs["file_format"] = "qpformat"
            h5.attrs["imaging_modality"] = "off-axis holography"

    def get_raw_data_iterator(self):
        for pp in sorted(self.path_raw.rglob("*.mat")):
            if (is_valid_h5_file(pp)
                and h5_file_contains(pp, "topogMap")
                and h5_file_contains(pp, "res")
                    and h5_file_contains(pp, "lambda")):
                yield [pp]

    def get_target_path(self, path_list):
        """Get the target path .h5 for a path_list"""
        target_mat = super(OAHRecipe, self).get_target_path(path_list)
        return target_mat.with_suffix(".h5")



import shutil
import warnings

from ..util import hashfile
from ..recipe import Recipe


class CatchAllRecipe(Recipe):
    """Copy all files, except known junk files"""

    def convert_dataset(self, path_list, temp_path, **kwargs):
        """Create a symlink and if that fails, copy the file"""
        try:
            temp_path.symlink_to(path_list[0])
        except BaseException:
            warnings.warn("Symbolic link generation failed, falling back "
                          + "to direct copying (which is slower)!")
            shutil.copy2(path_list[0], temp_path)
            # Perform a preliminary hash check to make sure that this
            # copy process was done properly.
            if hashfile(path_list[0]) != hashfile(temp_path):
                raise ValueError(
                    f"Initial hash verification failed for {path_list[0]}!")

    def get_raw_data_iterator(self):
        ignore_list = [
            ".DS_Store",
            "._.DS_Store",
            "Thumbs.db",
        ]
        for pp in self.path_raw.rglob("*"):
            if not pp.is_dir() and pp.name not in ignore_list:
                yield [pp]



"""
Structuring paths in a tree to efficiently handle files.
"""

import pathlib
import numpy as np
from PyQt6 import QtWidgets


class PathError(BaseException):
    pass


class PathTree:
    def __init__(self, path: pathlib.Path, depth_limit: int = 8):
        """
        Class for handling paths in a tree structure.

        Parameters
        ----------
        path : str or pathlib.Path
            The root of the path tree. Must be absolute path. If the path
            points to a file, the parent directory of that file is used.
        depth_limit : int
            Maximum depth of subdirectories that the PathTree object will load.
            Avoid large numbers on directories with many subfolders and files
            to reduce speed and performance issues.
        """
        if not isinstance(path, pathlib.Path):
            raise TypeError(f"Path for PathTree must be `pathlib.Path`. Got "
                            f"{type(path)} instead!")
        if path.is_dir():
            self.tree_root = path
        elif path.is_file():
            self.tree_root = path.parent
        else:
            raise PathError("Root path given as input for PathTree is neither "
                            "a valid directory nor a valid file. Don't know "
                            "what to do.")
        self.depth_limit = depth_limit
        self.children = {}

        for file_obj in self.tree_root.iterdir():
            if file_obj.is_dir() and self.depth_limit > 1:
                self.children[file_obj.name] = PathTree(file_obj,
                                                        self.depth_limit - 1)
        self.list_child_dirs = []
        if self.depth_limit == 1:
            # in case we are in the "deepest" PathTree, only make a list of
            # subdirectories, but not full PathTree objects. This will make
            # sure that the function `get_tree_depth` will always return
            # correct values
            for file_obj in self.tree_root.iterdir():
                if file_obj.is_dir():
                    self.list_child_dirs.append(file_obj.name)
        self.tree_depth = self.get_tree_depth()

    def __contains__(self, other_path: pathlib.Path) -> bool:
        # this will not work perfectly, since the tree might just not have
        # enough depth to contain the path in question (if self.depth_limit is
        # rather small).
        # To check the whole tree, use:
        # path in tree.retrieve_full_path_tree()
        if other_path.samefile(self.tree_root):
            return True
        elif self.children:
            for child_tree in self.children.values():
                return other_path in child_tree
        elif self.list_child_dirs:
            contained = False
            for sub_dir in self.list_child_dirs:
                if other_path.samefile(self.tree_root / sub_dir):
                    contained = True
                    break
            return contained
        else:
            return False

    def get_tree_depth(self) -> int:
        """Returns the depth of the path tree. If there are no subfolders,
        the depth is 1. Works recursively!"""
        if self.children:
            child_depths = [child.get_tree_depth() for child in
                            self.children.values()]
            return np.max(child_depths) + 1
        else:
            return 1

    def get_file_list(self) -> list[pathlib.Path]:
        """Return a list of all files in the root directory."""
        return [x for x in self.tree_root.glob("*.*") if
                x.is_file() and not x.name[0] == "."]

    def retrieve_full_path_tree(self):
        """Returns a `PathTree` object containing all files and subdirectories,
        without a depth limit (depth_limit=100)."""
        return PathTree(self.tree_root, depth_limit=100)


def list_items_in_tree(p_tree: PathTree,
                       tree_widget: QtWidgets.QTreeWidget,
                       h_level: int = 0,
                       depth_limit: int = 24) -> None:
    """Recursive function used for visualisation of the paths saved in a
    `PathTree` object.
    To avoid problems when working with many nested subdirectories, a limit
    for the maximum depth of subfolders is given by the parameter
    `depth_limit`.

    Parameters
    ----------
    p_tree : PathTree
        An instance of :class:`PathTree`, contains all the information about
        files and subfolders.
    tree_widget : QtWidgets.QTreeWidget
        The QtWidget used to display the files and subfolders in a certain
        directory (=the root directory of :PathTree:`p_tree`)
    h_level : int
        Counter for the hierarchy level, tells how deep into subfolders the
        algorithms went. Needed for displaying files in the right column of the
        QTreeWidget. Basically counts the recursion depth of the function.
    depth_limit : int
        Maximum depth of subfolders that will be displayed.
    """
    root_item = QtWidgets.QTreeWidgetItem(tree_widget)
    root_item.setText(h_level, p_tree.tree_root.name)
    if h_level < depth_limit:
        for file in p_tree.get_file_list():
            item = QtWidgets.QTreeWidgetItem(tree_widget)
            item.setText(h_level + 1, file.name)
    if p_tree.children and h_level < depth_limit:
        for child_tree in p_tree.children.values():
            list_items_in_tree(child_tree, tree_widget, h_level + 1,
                               depth_limit)
    if p_tree.list_child_dirs and h_level <= depth_limit:
        for sub_dir in p_tree.list_child_dirs:
            item = QtWidgets.QTreeWidgetItem(tree_widget)
            item.setText(h_level + 1, sub_dir)



# -----------------------------------------------------------------------------
# Copyright (c) 2019, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License with exception
# for distributing bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

# Hook for MPL-Data-Cast
from PyInstaller.utils.hooks import collect_data_files

# Data files
datas = collect_data_files("mpl_data_cast", include_py_files=True)
datas += collect_data_files("mpl_data_cast", subdir="gui/img")

# Add the Zstandard library used by dclab
datas += collect_data_files("hdf5plugin", includes=["plugins/libh5zstd.*"])



import inspect
import pathlib
import time

import click

from .. import recipe as mpldc_recipe


@click.group()
def cli():
    pass


@cli.command(short_help="List available recipes that can be used in the "
                        + "``mpldc cast`` subcommand")
def list_recipes():
    recipes = mpldc_recipe.get_available_recipe_names()
    col1len = max(len(r) for r in recipes) + 2
    for rec in recipes:
        cls = mpldc_recipe.map_recipe_name_to_class(rec)
        col1 = rec + (" " * (col1len - len(rec)))
        doc = cls.__doc__.split("\n")[0]
        click.secho(f"{col1} {doc}")


@cli.command(short_help="Convert and copy experimental data")
@click.argument("path_raw",
                type=click.Path(exists=True,
                                file_okay=False,
                                resolve_path=True,
                                path_type=pathlib.Path))
@click.argument("path_target",
                type=click.Path(file_okay=False,
                                writable=True,
                                resolve_path=True,
                                path_type=pathlib.Path))
@click.option("-r", "--recipe", type=str, default="CatchAll",
              help="specifies recipe to use, defaults to 'CatchAll'; see the "
                   + "``mpldc list-recipes`` subcommand for more options")
@click.option("-o", "--options", type=str, default=None,
              help="comma-separated keyword arguments passed to the recipe's "
                   + "`convert_dataset` method, e.g. "
                   + "wavelength=984e-9,pixel_size=1.2e-6")
def cast(path_raw, path_target, recipe="CatchAll", options=None):
    """Cast data from a source directory to a target directory

    This will convert all data under the tree in PATH_RAW and
    copy them to PATH_TARGET.
    """
    # get the actual class
    rcls = mpldc_recipe.map_recipe_name_to_class(recipe)
    # instantiate the class
    rp = rcls(path_raw, path_target)

    # get types of the recipes
    kwarg_dtypes = {}
    sig = inspect.signature(rp.convert_dataset)
    for p in sig.parameters:
        if p in ["path_list", "temp_path"]:
            continue
        kwarg_dtypes[p] = sig.parameters[p].annotation

    # parse custom parameters from `options`
    kwargs = {}
    if options:
        entries = options.split(",")
        for entr in entries:
            if "=" not in entr:
                raise ValueError(f"Invalid option string: '{entr}'!")
            key, valuestr = [en.strip() for en in entr.split("=", 1)]
            if key not in kwarg_dtypes:
                raise ValueError("Recipe `recipe` does not implement option "
                                 + f"'{key}'; available options: "
                                 + f"{sorted(kwarg_dtypes.keys())}")
            kwargs[key] = kwarg_dtypes[key](valuestr)
    click.secho(f"Using recipe {recipe}.", bold=True)
    with CLICallback() as path_callback:
        result = rp.cast(path_callback=path_callback, **kwargs)
    if result["success"]:
        click.secho("Success!", bold=True)
    else:
        click.secho("Errors encountered for the following files: ", bold=True)
        for path, _ in result["errors"]:
            click.echo(f" - {path}")
        if click.confirm('Should I dump the tracebacks to "mpldc-dump.txt"?'):
            text = ""
            for path, tb in result["errors"]:
                text += f"PATH {path}:\n{tb}\n\n"
            pathlib.Path("mpldc-dump.txt").write_text(text)


class CLICallback:
    def __init__(self):
        self.counter = 0
        self.size = 0
        self.prev_len = 0
        self.time_start = time.monotonic()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        rate = self.get_rate()
        self.print(f"Processed {self.counter} files (~{rate:.1f} MB/s).")
        print("")

    def __call__(self, path: pathlib.Path) -> None:
        self.counter += 1
        name = click.format_filename(path, shorten=True)
        message = f"Processing file {self.counter}: {name}"
        rate = self.get_rate()
        if rate:
            message += f" ({rate:.1f}MB/s)"
        self.size += path.stat().st_size
        self.print(message)

    def get_rate(self) -> float:
        curtime = time.monotonic()
        if curtime > self.time_start:
            return self.size / 1024**2 / (curtime - self.time_start)
        else:
            return 0

    def print(self, message: str) -> None:
        print(" " * self.prev_len, end="\r")
        print(message, end="\r")
        self.prev_len = len(message)
